import { EventEmitter } from './events/eventEmitter';
import { IRealRendererEvents } from '../../constants/events/RealRendererEvents';
import { GraphDimensions, Color, RealRendererOptions, Stroke, RealRendererSettings, BGType, RealExport, SVGSections } from '../../types/RealRendererTypes';
export * as RealRendererTypes from '../../types/RealRendererTypes';
export * from '../../constants/defaults/RealRendererDefaults';
import { undo, redo } from './undo';
/**
 * General Real Renderer with no specific purpose. Should be extended to use.
 */
export declare class RealRenderer<EventTypes extends IRealRendererEvents = IRealRendererEvents> extends EventEmitter<EventTypes> {
    svg: SVGSVGElement;
    svgSections: SVGSections;
    dimensions: GraphDimensions;
    scaleFactor: number;
    _offsetX: number;
    _offsetY: number;
    originalDimensions: GraphDimensions;
    strokes: Stroke[];
    settings: RealRendererSettings;
    _strokeIndex: number;
    bgColor: Color;
    bgType: BGType;
    drawsPerFrame: number;
    timeStep: number;
    time: number;
    _doRender: boolean;
    undo: typeof undo;
    redo: typeof redo;
    /**
     *
     * @param options
     * @param eventList
     */
    constructor(options: RealRendererOptions, eventList?: (keyof EventTypes)[]);
    protected _setSVG(): void;
    protected _setViewBox(dimensions: GraphDimensions, xOffset: number, yOffset: number): void;
    protected _drawFunc(time: number): this;
    protected _draw(): this;
    protected _addStroke(stroke: Stroke): void;
    protected _render(): void;
    protected _display(stroke: Stroke): void;
    /**
     * Attaches to a DOM SVG element to render to.
     * @param svg The SVG element to attach.
     * @param dimensions Dimensions of the graph.
     * @returns Self for chaining.
     */
    attach(svg: SVGSVGElement, dimensions: GraphDimensions): this;
    /**
     * Start rendering.
     * @returns Self for chaining.
     */
    startRender(): this;
    /**
     * Stop rendering.
     * @returns Self for chaining.
     */
    stopRender(): this;
    /**
     * Toggle rendering.
     * @returns Self for chaining.
     */
    toggleRender(): this;
    /**
     * Draw a certain number of frames.
     * @param numDraws Number of frames to draw.
     * @returns Self for chaining.
     */
    draw(numDraws?: number): this;
    /**
     * Scale/zoom the graph.
     * @param scaleFactor Factor to scale the graph by. Larger number zooms in.
     * @returns Self for chaining.
     */
    scale(scaleFactor: number): this;
    /**
     * Change the offsets of the graph (for panning)
     * @param xOffset Offset in the x-direction.
     * @param yOffset Offset in the y-direction.
     * @returns Self for chaining.
     */
    changeOffsets(xOffset: number, yOffset: number): this;
    /**
     * Export the data of the graph in a certain format that can be used to load the data later. Load using .importData().
     * @returns Data of the graph in a storable and loadable format.
     */
    exportData(): RealExport;
    /**
     * Import the data exported by .export() method.
     * @param data Data exported by .export()
     * @returns Self for chaining.
     */
    importData(data: RealExport): this;
    /**
     * Resets the internal time counter.
     * @returns Self for chaining.
     */
    resetTime(): this;
    /**
     * Resets everything regarding the graph.
     * @returns Self for chaining.
     */
    reset(): this;
    /**
     * Changes the background of the graph.
     * @param newBG New background.
     * @returns Self for chaining.
     */
    changeBackground(newBG: BGType): this;
}
